#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.constructutils import Ver
from m1n1.utils import *

Ver.set_version(u)

from m1n1.shell import run_shell

from m1n1.agx import AGX
from m1n1.agx.context import *

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")
#p.pmgr_adt_clocks_enable("/arm-io/pmp")

# [cpu0] [0xfffffe00124bf5c0] MMIO: R.4   0x204d14000 (sgx, offset 0xd14000) = 0x0
p.read32(0x204000000 + 0xd14000)
# [cpu0] [0xfffffe00124bf9a8] MMIO: W.4   0x204d14000 (sgx, offset 0xd14000) = 0x70001
p.write32(0x204000000 + 0xd14000, 0x70001)

#p.read32(0x204010258)

agx = AGX(u)

mon = RegMonitor(u, ascii=True, bufsize=0x8000000)
agx.mon = mon

sgx = agx.sgx_dev
mon.add(sgx.gpu_region_base, sgx.gpu_region_size, "contexts")
mon.add(sgx.gfx_shared_region_base, sgx.gfx_shared_region_size, "gfx-shared")
mon.add(sgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

#addr, size = sgx.get_reg(0)
#mon.add(addr + 0x600000, size - 0x600000, "sgx")

addr, size = u.adt["/arm-io/aic"].get_reg(0)
mon.add(addr, size, "aic")

def unswizzle(addr, w, h, psize, dump=None, grid=False):
    tw = 64
    th = 64
    ntx = (w + tw - 1) // 64
    nty = (h + th - 1) // 64
    data = iface.readmem(addr, ntx * nty * psize * tw * th)
    new_data = []
    for y in range(h):
        ty = y // th
        for x in range(w):
            tx = x // tw
            toff = tw * th * psize * (ty * ntx + tx)
            j = x & (tw - 1)
            i = y & (th - 1)
            off = (
                ((j & 1) << 0) | ((i & 1) << 1) |
                ((j & 2) << 1) | ((i & 2) << 2) |
                ((j & 4) << 2) | ((i & 4) << 3) |
                ((j & 8) << 3) | ((i & 8) << 4) |
                ((j & 16) << 4) | ((i & 16) << 5) |
                ((j & 32) << 5) | ((i & 32) << 6))
            r,g,b,a = data[toff + psize*off: toff + psize*(off+1)]
            if grid:
                if x % 64 == 0 or y % 64 == 0:
                    r,g,b,a = 255,255,255,255
                elif x % 32 == 0 or y % 32 == 0:
                    r,g,b,a = 128,128,128,255
            new_data.append(bytes([b, g, r, a]))
    data = b"".join(new_data)
    if dump:
        open(dump, "wb").write(data)
    iface.writemem(addr, data)

try:
    agx.start()

    ctx_id = 3
    buffer_mgr_slot = 2

    #agx.initdata.regionA.add_to_mon(mon)
    #agx.initdata.regionB.add_to_mon(mon)
    #agx.initdata.regionC.add_to_mon(mon)

    #agx.initdata.regionB.unk_170.add_to_mon(mon)
    #agx.initdata.regionB.unk_178.add_to_mon(mon)
    #agx.initdata.regionB.unk_180.add_to_mon(mon)
    #agx.initdata.regionB.unk_190.add_to_mon(mon)
    #agx.initdata.regionB.unk_198.add_to_mon(mon)
    ##agx.initdata.regionB.fwlog_ring2.add_to_mon(mon)
    #agx.initdata.regionB.hwdata_a.add_to_mon(mon)
    #agx.initdata.regionB.hwdata_b.add_to_mon(mon)
    mon.poll()

    #agx.asc.work_for(0.3)

    #p.write32(agx.initdata.regionC._paddr + 0x8900, 0xffffffff)
    #p.write32(agx.initdata.regionC._paddr + 0x8904, 0xffffffff)
    #mon.poll()

    agx.kick_firmware()
    #agx.asc.work_for(0.3)

    #mon.poll()

    ##### Initialize context and load data

    ctx = GPUContext(agx)
    ctx.bind(ctx_id)

    #p.read32(0x204000000 + 0xd14000)
    #p.write32(0x204000000 + 0xd14000, 0x70001)

    #base = "gpudata/1tri/"
    #base = "gpudata/mesa-flag/"
    base = "gpudata/bunny/"
    ctx.load_blob(0x1100000000, True, base + "mem_0_0.bin")
    ctx.load_blob(0x1100008000, True, base + "mem_8000_0.bin")
    ctx.load_blob(0x1100010000, True, base + "mem_10000_0.bin")
    ctx.load_blob(0x1100058000, True, base + "mem_58000_0.bin")
    ctx.load_blob(0x1100060000, True, base + "mem_60000_0.bin")
    #ctx.load_blob(0x1100068000, True, base + "mem_68000_0.bin")
    ctx.load_blob(0x1500000000, False, base + "mem_1500000000_0.bin")
    ctx.load_blob(0x1500048000, False, base + "mem_1500048000_0.bin")
    ctx.load_blob(0x15000d0000, False, base + "mem_15000d0000_0.bin")
    ctx.load_blob(0x1500158000, False, base + "mem_1500158000_0.bin")
    ctx.load_blob(0x15001e0000, False, base + "mem_15001e0000_0.bin")
    ctx.load_blob(0x15001e8000, False, base + "mem_15001e8000_0.bin")
    ctx.load_blob(0x15001f0000, False, base + "mem_15001f0000_0.bin")
    ctx.load_blob(0x15001f8000, False, base + "mem_15001f8000_0.bin")
    #ctx.load_blob(0x1500490000, False, base + "mem_1500490000_0.bin")
    #ctx.load_blob(0x1500518000, False, base + "mem_1500518000_0.bin")
    #color = ctx.buf_at(0x1500200000, False, 1310720, "Color", track=False)
    #depth = ctx.buf_at(0x1500348000, False, 1310720, "Depth", track=False)
    color = ctx.buf_at(0x1500200000, False, 2129920, "Color", track=False)
    depth = ctx.buf_at(0x1500410000, False, 2129920, "Depth", track=False)
    ctx.load_blob(0x1500620000, False, base + "mem_1500620000_0.bin", track=False)
    ctx.load_blob(0x1500890000, False, base + "mem_1500890000_0.bin", track=False)

    mon.poll()

    p.memset32(color._paddr, 0xdeadbeef, color._size)
    p.memset32(depth._paddr, 0xdeadbeef, depth._size)

    stencil = ctx.buf_at(0x1510410000, False, 2129920, "Stencil", track=False)

    width = 800
    height = 600

    width_a = align_up(width, 64)
    height_a = align_up(height, 64)

    depth_addr = depth._addr

    ##### Initialize buffer manager

    #buffer_mgr = GPUBufferManager(agx, ctx, 26)
    buffer_mgr = GPUBufferManager(agx, ctx, 8)

    ##### Initialize work queues

    wq_3d = GPU3DWorkQueue(agx, ctx)
    wq_ta = GPUTAWorkQueue(agx, ctx)

    ##### TA stamps

    #Message 1: DAG: Non Sequential Stamp Updates seen entryIdx 0x41 roots.dag 0x1 stampIdx 0x7 stampValue 0x4100 channel 0xffffffa000163f58 channelRingCommandIndex 0x1

    prev_stamp_value = 0x4000
    stamp_value = 0x4100

    # start?
    stamp_ta1 = agx.kshared.new(BarrierCounter, name="TA stamp 1")
    stamp_ta1.value = prev_stamp_value
    stamp_ta1.push()

    # complete?
    stamp_ta2 = agx.kobj.new(BarrierCounter, name="TA stamp 2")
    stamp_ta2.value = prev_stamp_value
    stamp_ta2.push()

    ##### 3D stamps

    # start?
    stamp_3d1 = agx.kshared.new(BarrierCounter, name="3D stamp 1")
    stamp_3d1.value = prev_stamp_value
    stamp_3d1.push()

    # complete?
    stamp_3d2 = agx.kobj.new(BarrierCounter, name="3D stamp 2")
    stamp_3d2.value = prev_stamp_value
    stamp_3d2.push()

    ##### Some kind of feedback/status buffer, GPU managed?

    event_control = agx.kobj.new(EventControl)
    event_control.event_count = agx.kobj.new(Int32ul, "Event Count")
    event_control.base_stamp = 0x15 #0
    event_control.unk_c = 0
    event_control.unk_10 = 0x50
    event_control.push()

    ##### TVB allocations / Tiler config

    tile_width = 32
    tile_height = 32
    tiles_x = ((width + tile_width - 1) // tile_width)
    tiles_y = ((height + tile_height - 1) // tile_height)
    tiles = tiles_x * tiles_y

    tile_blocks_x = (tiles_x + 15) // 16
    tile_blocks_y = (tiles_y + 15) // 16
    tile_blocks = tile_blocks_x * tile_blocks_y

    tiling_params = TilingParameters()
    tiling_params.size1 = 0x14 * tile_blocks
    tiling_params.unk_4 = 0x88
    tiling_params.unk_8 = 0x202
    tiling_params.x_max = width - 1
    tiling_params.y_max = height - 1
    tiling_params.tile_count = ((tiles_y-1) << 12) | (tiles_x-1)
    tiling_params.x_blocks = (12 * tile_blocks_x) | (tile_blocks_x << 12) | (tile_blocks_x << 20)
    tiling_params.y_blocks = (12 * tile_blocks_y) | (tile_blocks_y << 12) | (tile_blocks_y << 20)
    tiling_params.size2 = 0x10 * tile_blocks
    tiling_params.size3 = 0x20 * tile_blocks
    tiling_params.unk_24 = 0x100
    tiling_params.unk_28 = 0x8000

    tvb_something_size = 0x800 * tile_blocks
    tvb_something = ctx.uobj.new_buf(tvb_something_size, "TVB Something")

    tvb_tilemap_size = 0x800 * tile_blocks
    tvb_tilemap = ctx.uobj.new_buf(tvb_tilemap_size, "TVB Tilemap")

    tvb_heapmeta_size = 0x4000
    tvb_heapmeta = ctx.uobj.new_buf(tvb_heapmeta_size, "TVB Heap Meta")

    ##### Buffer stuff?

    # buffer related?
    buf_desc = agx.kobj.new(BufferThing)
    buf_desc.unk_0 = 0x0
    buf_desc.unk_8 = 0x0
    buf_desc.unk_10 = 0x0
    buf_desc.unkptr_18 = ctx.uobj.buf(0x80, "BufferThing.unkptr_18")
    buf_desc.unk_20 = 0x0
    buf_desc.bm_misc_addr = buffer_mgr.misc_obj._addr
    buf_desc.unk_2c = 0x0
    buf_desc.unk_30 = 0x0
    buf_desc.unk_38 = 0x0
    buf_desc.push()

    uuid_3d = 0x4000a14
    uuid_ta = 0x4000a15
    encoder_id = 0x30009fb

    ##### 3D barrier command

    ev_ta = 6
    ev_3d = 7

    barrier_cmd = agx.kobj.new(WorkCommandBarrier)
    barrier_cmd.stamp = stamp_ta2
    barrier_cmd.stamp_value1 = 0x4100
    barrier_cmd.stamp_value2 = 0x4100
    barrier_cmd.event = ev_ta
    barrier_cmd.uuid = uuid_3d


    #stamp.add_to_mon(mon)
    #stamp2.add_to_mon(mon)

    print(barrier_cmd)

    wq_3d.submit(barrier_cmd.push())

    ##### 3D execution

    wc_3d = agx.kobj.new(WorkCommand3D)
    wc_3d.context_id = ctx_id
    wc_3d.unk_8 = 0
    wc_3d.event_control = event_control
    wc_3d.buffer_mgr = buffer_mgr.info
    wc_3d.buf_thing = buf_desc
    wc_3d.unk_emptybuf_addr = agx.kobj.buf(0x100, "unk_emptybuf")
    wc_3d.tvb_tilemap = tvb_tilemap._addr
    wc_3d.unk_40 = 0x88
    wc_3d.unk_48 = 0x1
    wc_3d.tile_blocks_y = tile_blocks_y * 4
    wc_3d.tile_blocks_x = tile_blocks_x * 4
    wc_3d.unk_50 = 0x0
    wc_3d.unk_58 = 0x0
    wc_3d.uuid1 = 0x3b315cae
    wc_3d.uuid2 = 0x3b6c7b92
    wc_3d.unk_68 = 0x0
    wc_3d.tile_count = tiles

    wc_3d.unk_buf = WorkCommand1_UnkBuf()
    wc_3d.unk_word = BarrierCounter()
    wc_3d.unk_buf2 = WorkCommand1_UnkBuf2()
    wc_3d.unk_buf2.unk_0 = 0
    wc_3d.unk_buf2.unk_8 = 0
    wc_3d.unk_buf2.unk_10 = 1
    wc_3d.ts1 = Timestamp()
    wc_3d.ts2 = Timestamp()
    wc_3d.ts3 = Timestamp()
    wc_3d.unk_914 = 0
    wc_3d.unk_918 = 0
    wc_3d.unk_920 = 0
    wc_3d.unk_924 = 1

    # Structures embedded in WorkCommand3D
    if True:
        wc_3d.struct_1 = Start3DStruct1()
        wc_3d.struct_1.store_pipeline_addr = 0x14004 # CHECKED
        wc_3d.struct_1.unk_8 = 0x0
        wc_3d.struct_1.unk_c = 0x0
        wc_3d.struct_1.uuid1 = wc_3d.uuid1
        wc_3d.struct_1.uuid2 = wc_3d.uuid2
        wc_3d.struct_1.unk_18 = 0x0
        wc_3d.struct_1.tile_blocks_y = tile_blocks_y * 4
        wc_3d.struct_1.tile_blocks_x = tile_blocks_x * 4
        wc_3d.struct_1.unk_24 = 0x0
        wc_3d.struct_1.tile_counts = ((tiles_y-1) << 12) | (tiles_x-1)
        wc_3d.struct_1.unk_2c = 0x8
        wc_3d.struct_1.depth_clear_val1 = 1.0 # works
        wc_3d.struct_1.stencil_clear_val1 = 0x0
        wc_3d.struct_1.unk_38 = 0x0
        wc_3d.struct_1.unk_3c = 0x1
        wc_3d.struct_1.unk_40_padding = bytes(0xb0)
        wc_3d.struct_1.depth_bias_array = Start3DArrayAddr(0x1500158000)
        wc_3d.struct_1.scissor_array = Start3DArrayAddr(0x15000d0000)
        wc_3d.struct_1.unk_110 = 0x0
        wc_3d.struct_1.unk_118 = 0x0
        wc_3d.struct_1.unk_120 = [0] * 37
        wc_3d.struct_1.unk_reload_pipeline = Start3DStorePipelineBinding(0xffff8212, 0xfffffff4)
        wc_3d.struct_1.unk_258 = 0
        wc_3d.struct_1.unk_260 = 0
        wc_3d.struct_1.unk_268 = 0
        wc_3d.struct_1.unk_270 = 0
        wc_3d.struct_1.reload_pipeline = Start3DClearPipelineBinding(0xffff8212, 0x13004) # CHECKED
        wc_3d.struct_1.depth_flags = 0x00000
        wc_3d.struct_1.unk_290 = 0x0
        wc_3d.struct_1.depth_buffer_ptr1 = depth_addr
        wc_3d.struct_1.unk_2a0 = 0x0
        wc_3d.struct_1.unk_2a8 = 0x0
        wc_3d.struct_1.depth_buffer_ptr2 = depth_addr
        wc_3d.struct_1.depth_buffer_ptr3 = depth_addr
        wc_3d.struct_1.unk_2c0 = 0x0
        wc_3d.struct_1.stencil_buffer_ptr1 = stencil._addr
        wc_3d.struct_1.unk_2d0 = 0x0
        wc_3d.struct_1.unk_2d8 = 0x0
        wc_3d.struct_1.stencil_buffer_ptr2 = stencil._addr
        wc_3d.struct_1.stencil_buffer_ptr3 = stencil._addr
        wc_3d.struct_1.unk_2f0 = [0x0, 0x0, 0x0]
        wc_3d.struct_1.aux_fb_unk0 = 0x4
        wc_3d.struct_1.unk_30c = 0x0
        wc_3d.struct_1.aux_fb = AuxFBInfo(0xc000, 0, width, height)
        wc_3d.struct_1.unk_320_padding = bytes(0x10)
        wc_3d.struct_1.unk_partial_store_pipeline = Start3DStorePipelineBinding(0xffff8212, 0xfffffff4)
        wc_3d.struct_1.partial_store_pipeline = Start3DStorePipelineBinding(0x12, 0x14004) # CHECKED
        wc_3d.struct_1.depth_clear_val2 = 1.0
        wc_3d.struct_1.stencil_clear_val2 = 0x0
        wc_3d.struct_1.context_id = ctx_id
        wc_3d.struct_1.unk_376 = 0x0
        wc_3d.struct_1.unk_378 = 0x8
        wc_3d.struct_1.unk_37c = 0x0
        wc_3d.struct_1.unk_380 = 0x0
        wc_3d.struct_1.unk_388 = 0x0
        wc_3d.struct_1.depth_dimensions = 0x12b831f #0xef827f

    if True:
        wc_3d.struct_2 = Start3DStruct2()
        wc_3d.struct_2.unk_0 = 0xa000
        wc_3d.struct_2.clear_pipeline = Start3DClearPipelineBinding(0xffff8002, 0x12004)
        wc_3d.struct_2.unk_18 = 0x88
        wc_3d.struct_2.scissor_array = 0x15000d0000
        wc_3d.struct_2.depth_bias_array = 0x1500158000
        wc_3d.struct_2.aux_fb =  wc_3d.struct_1.aux_fb
        wc_3d.struct_2.depth_dimensions = wc_3d.struct_1.depth_dimensions
        wc_3d.struct_2.unk_48 = 0x0
        wc_3d.struct_2.depth_flags = wc_3d.struct_1.depth_flags
        wc_3d.struct_2.depth_buffer_ptr1 = depth_addr
        wc_3d.struct_2.depth_buffer_ptr2 = depth_addr
        wc_3d.struct_2.stencil_buffer_ptr1 = stencil._addr
        wc_3d.struct_2.stencil_buffer_ptr2 = stencil._addr
        wc_3d.struct_2.unk_68 = [0] * 12
        wc_3d.struct_2.tvb_tilemap = tvb_tilemap._addr
        wc_3d.struct_2.tvb_heapmeta_addr = tvb_heapmeta._addr
        wc_3d.struct_2.unk_e8 = 0x50000000 * tile_blocks
        wc_3d.struct_2.tvb_heapmeta_addr2 = tvb_heapmeta._addr
        wc_3d.struct_2.unk_f8 = 0x10280 # TODO: varies 0, 0x280, 0x10000, 0x10280
        wc_3d.struct_2.aux_fb_ptr = 0x1500006000
        wc_3d.struct_2.unk_108 = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0]
        wc_3d.struct_2.pipeline_base = 0x1100000000
        wc_3d.struct_2.unk_140 = 0x8c60
        wc_3d.struct_2.unk_148 = 0x0
        wc_3d.struct_2.unk_150 = 0x0
        wc_3d.struct_2.unk_158 = 0x1c
        wc_3d.struct_2.unk_160_padding = bytes(0x1e8)

    if True:
        wc_3d.struct_6 = Start3DStruct6()
        wc_3d.struct_6.unk_0 = 0x0
        wc_3d.struct_6.unk_8 = 0x0
        wc_3d.struct_6.unk_10 = 0x0
        wc_3d.struct_6.encoder_id = encoder_id
        wc_3d.struct_6.unk_1c = 0xffffffff
        wc_3d.struct_6.unknown_buffer = 0x150000e000
        wc_3d.struct_6.unk_28 = 0x0
        wc_3d.struct_6.unk_30 = 0x1
        wc_3d.struct_6.unk_34 = 0x1

    if True:
        wc_3d.struct_7 = Start3DStruct7()
        wc_3d.struct_7.unk_0 = 0x0
        wc_3d.struct_7.stamp1 = stamp_3d1
        wc_3d.struct_7.stamp2 = stamp_3d2
        wc_3d.struct_7.stamp_value = stamp_value
        wc_3d.struct_7.ev_3d = ev_3d
        wc_3d.struct_7.unk_20 = 0x0
        wc_3d.struct_7.unk_24 = 0x0 # check
        wc_3d.struct_7.uuid = uuid_3d
        wc_3d.struct_7.prev_stamp_value = 0x0
        wc_3d.struct_7.unk_30 = 0x0

    wc_3d.set_addr() # Update inner structure addresses
    print("WC3D", hex(wc_3d._addr))
    print(" s1", hex(wc_3d.struct_1._addr))
    print(" s2", hex(wc_3d.struct_2._addr))
    print(" s6", hex(wc_3d.struct_6._addr))
    print(" s7", hex(wc_3d.struct_7._addr))

    ms = GPUMicroSequence(agx)

    start_3d = Start3DCmd()
    start_3d.struct1 = wc_3d.struct_1
    start_3d.struct2 = wc_3d.struct_2
    start_3d.buf_thing = buf_desc
    start_3d.unkptr_1c = agx.initdata.regionB.unkptr_178 + 8
    start_3d.unkptr_24 = wc_3d.unk_word._addr
    start_3d.struct6 = wc_3d.struct_6
    start_3d.struct7 = wc_3d.struct_7
    start_3d.cmdqueue_ptr = wq_3d.info._addr
    start_3d.workitem_ptr = wc_3d._addr
    start_3d.context_id = ctx_id
    start_3d.unk_50 = 0x1
    start_3d.unk_54 = 0x0
    start_3d.unk_58 = 0x2
    start_3d.unk_5c = 0x0
    start_3d.prev_stamp_value = 0x0
    start_3d.unk_68 = 0x0
    start_3d.unk_buf_ptr = wc_3d.unk_buf._addr
    start_3d.unk_buf2_ptr = wc_3d.unk_buf2._addr
    start_3d.unk_7c = 0x0
    start_3d.unk_80 = 0x0
    start_3d.unk_84 = 0x0
    start_3d.uuid = uuid_3d
    start_3d.attachments = [
        Attachment(color._addr, 0x2800, 0x10017),
        Attachment(depth._addr, 0x4100, 0x10017),
    ] + [Attachment(0, 0, 0)] * 14
    start_3d.num_attachments = 2
    start_3d.unk_190 = 0x0

    ms.append(start_3d)

    ts1 = TimestampCmd()
    ts1.unk_1 = 0x0
    ts1.unk_2 = 0x0
    ts1.unk_3 = 0x80
    ts1.ts0_addr = wc_3d.ts1._addr
    ts1.ts1_addr = wc_3d.ts2._addr
    ts1.ts2_addr = wc_3d.ts2._addr
    ts1.cmdqueue_ptr = wq_3d.info._addr
    ts1.unk_24 = 0x0
    ts1.uuid = uuid_3d
    ts1.unk_30_padding = 0x0
    ms.append(ts1)

    ms.append(WaitForInterruptCmd(0, 1, 0))

    ts2 = TimestampCmd()
    ts2.unk_1 = 0x0
    ts2.unk_2 = 0x0
    ts2.unk_3 = 0x0
    ts2.ts0_addr = wc_3d.ts1._addr
    ts2.ts1_addr = wc_3d.ts2._addr
    ts2.ts2_addr = wc_3d.ts3._addr
    ts2.cmdqueue_ptr = wq_3d.info._addr
    ts2.unk_24 = 0x0
    ts2.uuid = uuid_3d
    ts2.unk_30_padding = 0x0
    ms.append(ts2)

    finish_3d = Finalize3DCmd()
    finish_3d.uuid = uuid_3d
    finish_3d.unk_8 = 0
    finish_3d.stamp = stamp_3d2
    finish_3d.stamp_value = stamp_value
    finish_3d.unk_18 = 0
    finish_3d.buf_thing = buf_desc
    finish_3d.buffer_mgr = buffer_mgr.info
    finish_3d.unk_2c = 1
    finish_3d.unkptr_34 = agx.initdata.regionB.unkptr_178 + 8
    finish_3d.struct7 = wc_3d.struct_7
    finish_3d.unkptr_44 = wc_3d.unk_word._addr
    finish_3d.cmdqueue_ptr = wq_3d.info._addr
    finish_3d.workitem_ptr = wc_3d._addr
    finish_3d.unk_5c = ctx_id
    finish_3d.unk_buf_ptr = wc_3d.unk_buf._addr
    finish_3d.unk_6c = 0
    finish_3d.unk_74 = 0
    finish_3d.unk_7c = 0
    finish_3d.unk_84 = 0
    finish_3d.unk_8c = 0
    finish_3d.startcmd_offset = -0x200
    finish_3d.unk_98 = 1
    ms.append(finish_3d)
    ms.finalize()

    wc_3d.microsequence_ptr = ms.obj._addr
    wc_3d.microsequence_size = ms.size

    print(wc_3d)

    wc_3d.push()
    ms.dump()
    print(wc_3d)
    wq_3d.submit(wc_3d)

    ##### TA init

    #print(ctx_info)

    wc_initbm = agx.kobj.new(WorkCommandInitBM)
    wc_initbm.context_id = ctx_id
    wc_initbm.unk_8 = buffer_mgr_slot
    wc_initbm.unk_c = 0
    wc_initbm.unk_10 = buffer_mgr.info.block_count
    wc_initbm.buffer_mgr = buffer_mgr.info
    wc_initbm.stamp_value = stamp_value
    wc_initbm.push()

    print(wc_initbm)
    wq_ta.submit(wc_initbm)

    ##### TA execution

    wc_ta = agx.kobj.new(WorkCommandTA)
    wc_ta.context_id = ctx_id
    wc_ta.unk_8 = 0
    wc_ta.event_control = event_control
    wc_ta.unk_14 = buffer_mgr_slot
    wc_ta.buffer_mgr = buffer_mgr.info
    wc_ta.buf_thing = buf_desc
    wc_ta.unk_emptybuf_addr = wc_3d.unk_emptybuf_addr
    wc_ta.unk_34 = 0x0

    wc_ta.unk_154 = bytes(0x268)
    wc_ta.unk_3e8 = bytes(0x74)
    wc_ta.unk_594 = WorkCommand0_UnkBuf()

    wc_ta.ts1 = Timestamp()
    wc_ta.ts2 = Timestamp()
    wc_ta.ts3 = Timestamp()
    wc_ta.unk_5c4 = 0
    wc_ta.unk_5c8 = 0
    wc_ta.unk_5cc = 0
    wc_ta.unk_5d0 = 0
    wc_ta.unk_5d4 = 0x27 #1

    # Structures embedded in WorkCommandTA
    if True:

        wc_ta.tiling_params = tiling_params
        #wc_ta.tiling_params.unk_0 = 0x28
        #wc_ta.tiling_params.unk_4 = 0x88
        #wc_ta.tiling_params.unk_8 = 0x202
        #wc_ta.tiling_params.x_max = 639
        #wc_ta.tiling_params.y_max = 479
        #wc_ta.tiling_params.unk_10 = 0xe013
        #wc_ta.tiling_params.unk_14 = 0x20_20_18
        #wc_ta.tiling_params.unk_18 = 0x10_10_0c
        #wc_ta.tiling_params.unk_1c = 0x20
        #wc_ta.tiling_params.unk_20 = 0x40
        #wc_ta.tiling_params.unk_24 = 0x100
        #wc_ta.tiling_params.unk_28 = 0x8000

    if True:
        wc_ta.struct_2 = StartTACmdStruct2()
        wc_ta.struct_2.unk_0 = 0x200
        wc_ta.struct_2.unk_8 = 0x1e3ce508 # fixed
        wc_ta.struct_2.unk_c = 0x1e3ce508 # fixed
        wc_ta.struct_2.tvb_tilemap = tvb_tilemap._addr
        wc_ta.struct_2.unkptr_18 = 0x0
        wc_ta.struct_2.unkptr_20 = tvb_something._addr
        wc_ta.struct_2.tvb_heapmeta_addr = tvb_heapmeta._addr | 0x8000000000000000
        wc_ta.struct_2.iogpu_unk_54 = 0x6b0003 # fixed
        wc_ta.struct_2.iogpu_unk_55 = 0x3a0012 # fixed
        wc_ta.struct_2.iogpu_unk_56 = 0x1 # fixed
        wc_ta.struct_2.unk_40 = 0x0 # fixed
        wc_ta.struct_2.unk_48 = 0xa000 # fixed
        wc_ta.struct_2.unk_50 = 0x88 # fixed
        wc_ta.struct_2.tvb_heapmeta_addr2 = tvb_heapmeta._addr
        wc_ta.struct_2.unk_60 = 0x0 # fixed
        wc_ta.struct_2.unk_68 = 0x0 # fixed
        wc_ta.struct_2.iogpu_deflake_1 = 0x15000052a0
        wc_ta.struct_2.iogpu_deflake_2 = 0x1500005020
        wc_ta.struct_2.unk_80 = 0x1 # fixed
        wc_ta.struct_2.iogpu_deflake_3 = 0x1500005000
        wc_ta.struct_2.encoder_addr = 0x1500048000
        wc_ta.struct_2.unk_98 = [0x0, 0x0] # fixed
        wc_ta.struct_2.unk_a8 = 0xa041 # fixed
        wc_ta.struct_2.unk_b0 = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0] # fixed
        wc_ta.struct_2.pipeline_base = 0x1100000000
        wc_ta.struct_2.unk_e8 = 0x0 # fixed
        wc_ta.struct_2.unk_f0 = 0x1c # fixed
        wc_ta.struct_2.unk_f8 = 0x8c60 # fixed
        wc_ta.struct_2.unk_100 = [0x0, 0x0, 0x0] # fixed
        wc_ta.struct_2.unk_118 = 0x1c # fixed

    if True:
        wc_ta.struct_3 = StartTACmdStruct3()
        wc_ta.struct_3.unk_480 = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0] # fixed
        wc_ta.struct_3.unk_498 = 0x0 # fixed
        wc_ta.struct_3.unk_4a0 = 0x0 # fixed
        wc_ta.struct_3.iogpu_deflake_1 = 0x15000052a0
        wc_ta.struct_3.unk_4ac = 0x0 # fixed
        wc_ta.struct_3.unk_4b0 = 0x0 # fixed
        wc_ta.struct_3.unk_4b8 = 0x0 # fixed
        wc_ta.struct_3.unk_4bc = 0x0 # fixed
        wc_ta.struct_3.unk_4c4_padding = bytes(0x48)
        wc_ta.struct_3.unk_50c = 0x0 # fixed
        wc_ta.struct_3.unk_510 = 0x0 # fixed
        wc_ta.struct_3.unk_518 = 0x0 # fixed
        wc_ta.struct_3.unk_520 = 0x0 # fixed
        wc_ta.struct_3.unk_528 = 0x0 # fixed
        wc_ta.struct_3.unk_52c = 0x0 # fixed
        wc_ta.struct_3.unk_530 = 0x0 # fixed
        wc_ta.struct_3.encoder_id = encoder_id
        wc_ta.struct_3.unk_538 = 0x0 # fixed
        wc_ta.struct_3.unk_53c = 0xffffffff
        wc_ta.struct_3.unknown_buffer = wc_3d.struct_6.unknown_buffer
        wc_ta.struct_3.unk_548 = 0x0 # fixed
        wc_ta.struct_3.unk_550 = [
            0x0, 0x0, # fixed
            0x0, # 1 for boot stuff?
            0x0, 0x0, 0x0] # fixed
        wc_ta.struct_3.stamp1 = stamp_ta1
        wc_ta.struct_3.stamp2 = stamp_ta2
        wc_ta.struct_3.stamp_value = stamp_value
        wc_ta.struct_3.ev_ta = ev_ta
        wc_ta.struct_3.unk_580 = 0x0 # fixed
        wc_ta.struct_3.unk_584 = 0x0 # 1 for boot stuff?
        wc_ta.struct_3.uuid2 = uuid_ta
        #wc_ta.struct_3.unk_58c = [0x0, 0x0]
        wc_ta.struct_3.unk_58c = [0x1, 0x0]

    wc_ta.set_addr() # Update inner structure addresses
    #print("wc_ta", wc_ta)

    ms = GPUMicroSequence(agx)

    start_ta = StartTACmd()
    start_ta.tiling_params = wc_ta.tiling_params
    start_ta.struct2 = wc_ta.struct_2
    start_ta.buffer_mgr = buffer_mgr.info
    start_ta.buf_thing = buf_desc
    start_ta.unkptr_24 = agx.initdata.regionB.unkptr_170 + 4
    start_ta.cmdqueue_ptr = wq_ta.info._addr
    start_ta.context_id = ctx_id
    start_ta.unk_38 = 1
    start_ta.unk_3c = 1 #0
    start_ta.unk_40 = buffer_mgr_slot
    start_ta.unk_48 = 1 #0
    start_ta.unk_50 = 0
    start_ta.struct3 = wc_ta.struct_3

    start_ta.unkptr_5c = wc_ta.unk_594._addr
    start_ta.unk_64 = 0x0 # fixed
    start_ta.uuid = uuid_ta
    start_ta.unk_70 = 0x0 # fixed
    start_ta.unk_74 = [ # fixed
        0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
        0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
    ]
    start_ta.unk_15c = 0x0 # fixed
    start_ta.unk_160 = 0x0 # fixed
    start_ta.unk_168 = 0x0 # fixed
    start_ta.unk_16c = 0x0 # fixed
    start_ta.unk_170 = 0x0 # fixed
    start_ta.unk_178 = 0x0 # fixed
    ms.append(start_ta)

    ts1 = TimestampCmd()
    ts1.unk_1 = 0x0
    ts1.unk_2 = 0x0
    ts1.unk_3 = 0x80
    ts1.ts0_addr = wc_ta.ts1._addr
    ts1.ts1_addr = wc_ta.ts2._addr
    ts1.ts2_addr = wc_ta.ts2._addr
    ts1.cmdqueue_ptr = wq_ta.info._addr
    ts1.unk_24 = 0x0
    ts1.uuid = uuid_ta
    ts1.unk_30_padding = 0x0
    ms.append(ts1)

    ms.append(WaitForInterruptCmd(1, 0, 0))

    ts2 = TimestampCmd()
    ts2.unk_1 = 0x0
    ts2.unk_2 = 0x0
    ts2.unk_3 = 0x0
    ts2.ts0_addr = wc_ta.ts1._addr
    ts2.ts1_addr = wc_ta.ts2._addr
    ts2.ts2_addr = wc_ta.ts3._addr
    ts2.cmdqueue_ptr = wq_ta.info._addr
    ts2.unk_24 = 0x0
    ts2.uuid = uuid_ta
    ts2.unk_30_padding = 0x0
    ms.append(ts2)

    finish_ta = FinalizeTACmd()
    finish_ta.buf_thing = buf_desc
    finish_ta.buffer_mgr = buffer_mgr.info
    finish_ta.unkptr_14 = agx.initdata.regionB.unkptr_170 + 4
    finish_ta.cmdqueue_ptr = wq_ta.info._addr
    finish_ta.context_id = ctx_id
    finish_ta.unk_28 = 0x0 # fixed
    finish_ta.struct3 = wc_ta.struct_3
    finish_ta.unk_34 = 0x0 # fixed
    finish_ta.uuid = uuid_ta
    finish_ta.stamp = stamp_ta2
    finish_ta.stamp_value = stamp_value
    finish_ta.unk_48 = 0x0 # fixed
    finish_ta.unk_50 = 0x0 # fixed
    finish_ta.unk_54 = 0x0 # fixed
    finish_ta.unk_58 = 0x0 # fixed
    finish_ta.unk_60 = 0x0 # fixed
    finish_ta.unk_64 = 0x0 # fixed
    finish_ta.unk_68 = 0x0 # fixed
    finish_ta.startcmd_offset = -0x1e8 # fixed
    finish_ta.unk_70 = 0x0 # fixed
    ms.append(finish_ta)

    ms.finalize()

    wc_ta.unkptr_45c = tvb_something._addr
    wc_ta.tvb_size = tvb_something_size
    wc_ta.microsequence_ptr = ms.obj._addr
    wc_ta.microsequence_size = ms.size
    wc_ta.ev_3d = ev_3d
    wc_ta.stamp_value = stamp_value

    wc_ta.push()
    ms.dump()

    mon.poll()

    print(wc_ta)
    wq_ta.submit(wc_ta)

    ##### Run queues
    agx.ch.queue[2].q_3D.run(wq_3d, ev_3d)
    agx.ch.queue[2].q_TA.run(wq_ta, ev_ta)

    ##### Wait for work
    agx.asc.work_for(0.3)
    print("3D:")
    print(wq_3d.info.pull())
    print("TA:")
    print(wq_ta.info.pull())
    print("Barriers:")
    print(stamp_ta1.pull())
    print(stamp_ta2.pull())
    print(stamp_3d1.pull())
    print(stamp_3d2.pull())

    event_control.pull()
    print(event_control)

    print("==")
    mon.poll()
    print("==")
    #agx.kick_firmware()
    agx.asc.work_for(0.3)
    p.read32(0x204000000 + 0xd14000)
    # [cpu0] [0xfffffe00124bf9a8] MMIO: W.4   0x204d14000 (sgx, offset 0xd14000) = 0x70001
    p.write32(0x204000000 + 0xd14000, 0x70001)

    #agx.uat.dump(ctx_id)

    fault_code = p.read64(0x204017030)
    fault_addr = fault_code >> 24
    if fault_addr & 0x8000000000:
        fault_addr |= 0xffffff8000000000
    print(f"FAULT CODE: {fault_code:#x}")
    base, obj = agx.find_object(fault_addr)
    if obj is not None:
        print(f"Faulted at : {fault_addr:#x}: {obj!s} + {fault_addr - base:#x}")
    #agx.kick_firmware()
    mon.poll()

    #print(buffer_mgr.info.pull())
    #print(buffer_mgr.counter_obj.pull())
    #print(buffer_mgr.misc_obj.pull())
    #print(buffer_mgr.block_ctl_obj.pull())

    width = 800
    height = 600

    unswizzle(color._paddr, width, height, 4, "fb.bin", grid=True)
    unswizzle(depth._paddr, width, height, 4, "depth.bin", grid=True)

    p.fb_blit(0, 0, width, height, color._paddr, width)

    print("TVB something:")
    chexdump(iface.readmem(tvb_something._paddr, tvb_something._size), stride=16, abbreviate=False)

    print("TVB list:")
    chexdump(iface.readmem(tvb_tilemap._paddr, tvb_tilemap._size), stride=5, abbreviate=False)

    print("Tile params:")
    print(f"X: {tiles_x} ({tile_blocks_x})")
    print(f"Y: {tiles_y} ({tile_blocks_y})")
    print(f"Total: {tiles} ({tile_blocks})")

    agx.stop()
except:
    #agx.uat.dump(ctx_id)
    p.reboot()
    raise
    #agx.stop()

#time.sleep(10)
#p.reboot()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.pmu import PMU

PMU(u).reset_panic_counter()

p.reboot()



# SPDX-License-Identifier: MIT
import os, struct, sys, time

from .hv import HV
from .proxy import *
from .proxyutils import *
from .sysreg import *
from .tgtypes import *
from .utils import *
from .hw.pmu import PMU

# Create serial connection
iface = UartInterface()
# Construct m1n1 proxy layer over serial connection
p = M1N1Proxy(iface, debug=False)
# Customise parameters of proxy and serial port
# based on information sent over the connection
bootstrap_port(iface, p)

# Initialise the Proxy interface from values fetched from
# the remote end
u = ProxyUtils(p)
# Build a Register Monitoring object on Proxy Interface
mon = RegMonitor(u)
hv = HV(iface, p, u)

fb = u.ba.video.base

PMU(u).reset_panic_counter()

print(f"m1n1 base: 0x{u.base:x}")

PMU(u).reset_panic_counter()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.i2c import I2C, I2CRegMapDev
from m1n1.hw.codecs.cs42l84 import *

class CS42L84(I2CRegMapDev):
    REGMAP = CS42L84Regs
    ADDRESSING = (0, 2)

def read_devid():
    pass

def sense_Z():
    r = l84.regs

    r.HS_CLAMP_DISABLE.set(HS_CLAMP_DISABLE=1)
    r.DAC_CTRL2.set(PULLDOWN_R=E_PULLDOWN_R.R_1K1OHMS)
    r.DCID_CTRL1.set(Z_RANGE=E_DCID_Z_RANGE.UNK2)
    r.DCID_CTRL2.set(GND_SEL=E_DCID_GND_SEL.HS3)
    r.MSM_BLOCK_EN3.set(DCID_EN=1)
    r.DCID_CTRL3.set(START=0)
    r.DCID_CTRL3.set(START=1)

    while not r.DCID_STATUS.reg.DONE:
        pass

    reading = r.DCID_STATUS.reg.OVERALL
    offset_trim = r.DCID_TRIM_OFFSET.val - 128
    slope_trim = r.DCID_TRIM_SLOPE.val - 128
    pulldown_trim = r.DCID_PULLDOWN_TRIM.val - 128

    Y_overall = ((reading + 0.5) * 0.01086 - 1.0 / (1 - slope_trim * 0.001375)) \
        / (614.0 + offset_trim * 0.125)
    Y_pulldown = 1.0 / (1100 - pulldown_trim*2)

    if Y_overall > Y_pulldown:
        Z_headphones = 1.0 / (Y_overall - Y_pulldown)
    else:
        Z_headphones = float('inf')

    r.MSM_BLOCK_EN3.set(DCID_EN=0)
    r.DAC_CTRL2.set(PULLDOWN_R=E_PULLDOWN_R.NONE)

    return Z_headphones

def init_ring_tip_sense():
    l84.regs.MIC_DET_CTRL4.set(LATCH_TO_VP=1)
    l84.regs.TIP_SENSE_CTRL2.set(CTRL=E_TIP_SENSE_CTRL.SHORT_DET)

    l84.regs.RING_SENSE_CTRL.set(INV=1, UNK1=1,
        RISETIME=E_DEBOUNCE_TIME.T_125MS, FALLTIME=E_DEBOUNCE_TIME.T_125MS)
    l84.regs.TIP_SENSE_CTRL.set(INV=1,
        RISETIME=E_DEBOUNCE_TIME.T_500MS, FALLTIME=E_DEBOUNCE_TIME.T_125MS)
    l84.regs.MSM_BLOCK_EN3.set(TR_SENSE_EN=1)

def wait_for_plug():
    while not l84.regs.TR_SENSE_STATUS.reg.TIP_PLUG:
        time.sleep(0.001)

def wait_for_unplug():
    while l84.regs.TR_SENSE_STATUS.reg.TIP_UNPLUG:
        time.sleep(0.001)


p.pmgr_adt_clocks_enable("/arm-io/i2c2")
i2c2 = I2C(u, "/arm-io/i2c2")

p.write32(0x2921f0010, 0x76a02) # invoke reset
p.write32(0x2921f0010, 0x76a03) # out of reset

l84 = CS42L84(i2c2, 0x4b)

init_ring_tip_sense()

while True:
    print("Waiting for plug... ", end=""); sys.stdout.flush()
    wait_for_plug()

    print("measuring... ", end=""); sys.stdout.flush()
    print(f"{sense_Z():.1f} ohms... ", end=""); sys.stdout.flush()

    wait_for_unplug()
    print("yanked")



# SPDX-License-Identifier: MIT
import struct
from construct import *
from copy import deepcopy

def round_up(x, y): return ((x + (y - 1)) & (-y))
def round_down(x, y): return (x - (x % y))

AOPBootargsItem = Struct(
    "key" / PaddedString(4, "utf8"),
    "size" / Int32ul,
)

class AOPBootargs:
    def __init__(self, bytes_):
        self.blob = bytearray(bytes_)
        self.index = self.build_index(self.blob)

    def build_index(self, blob):
        off = 0
        fields = []
        while off < len(blob):
            item = AOPBootargsItem.parse(blob[off:off+AOPBootargsItem.sizeof()])
            off += AOPBootargsItem.sizeof()
            fields.append((item.key, (off, item.size)))
            off += item.size
        if off > len(blob):
            raise ValueError('blob overran during parsing')
        return dict(fields)

    def items(self):
        for key, span in self.index.items():
            off, length = span
            yield key, self.blob[off:off + length]

    def __getitem__(self, key):
        off, length = self.index[key]
        return bytes(self.blob[off:off + length])

    def __setitem__(self, key, value):
        off, length = self.index[key]
        if type(value) is int:
            value = int.to_bytes(value, length, byteorder='little')
        elif type(value) is str:
            value = value.encode('ascii')
        if len(value) > length:
            raise ValueError(f'field {key:s} overflown')
        self.blob[off:off + length] = value

    def update(self, keyvals):
        for key, val in keyvals.items():
            self[key] = val

    def keys(self):
        return self.index.keys()

    def dump(self, logger):
        for key, val in self.items():
            logger(f"{key:4s} = {val}")

    def dump_diff(self, other, logger):
        assert self.index == other.index
        for key in self.keys():
            if self[key] != other[key]:
                logger(f"\t{key:4s} = {self[key]} -> {other[key]}")

    def to_bytes(self):
        return bytes(self.blob)

class AOPBase:
    def __init__(self, u):
        self.u = u
        self.nub_base = u.adt["/arm-io/aop/iop-aop-nub"].region_base
        if u.adt["arm-io"].compatible[0] == "arm-io,t6000":
            # argh
            self.nub_base -= 0x2_0000_0000

    @property
    def _bootargs_span(self):
        """
        [cpu1] MMIO: R.4   0x24ac0022c (aop[2], offset 0x22c) = 0xaffd8 // offset
        [cpu1] MMIO: R.4   0x24ac00230 (aop[2], offset 0x230) = 0x2ae // size
        [cpu1] MMIO: R.4   0x24ac00234 (aop[2], offset 0x234) = 0x82000 // va? low
        [cpu1] MMIO: R.4   0x24ac00238 (aop[2], offset 0x238) = 0x0 // va? high
        [cpu1] MMIO: R.4   0x24ac0023c (aop[2], offset 0x23c) = 0x4ac82000 // phys low
        [cpu1] MMIO: R.4   0x24ac00240 (aop[2], offset 0x240) = 0x2 // phys high
        [cpu1] MMIO: W.4   0x24acaffd8 (aop[2], offset 0xaffd8) = 0x53544b47 // start of bootargs
        [cpu1] MMIO: W.4   0x24acaffdc (aop[2], offset 0xaffdc) = 0x8
        [cpu1] MMIO: W.4   0x24acaffe0 (aop[2], offset 0xaffe0) = 0x73eed2a3
        ...
        [cpu1] MMIO: W.4   0x24acb0280 (aop[2], offset 0xb0280) = 0x10000
        [cpu1] MMIO: W.4   0x24acb0284 (aop[2], offset 0xb0284) = 0x0 // end of bootargs
        """
        offset = self.u.proxy.read32(self.nub_base + 0x22c) # 0x224 in 12.3
        size = self.u.proxy.read32(self.nub_base + 0x230) # 0x228 in 12.3
        return (self.nub_base + offset, size)

    def read_bootargs(self):
        addr, size = self._bootargs_span
        blob = self.u.proxy.iface.readmem(addr, size)
        return AOPBootargs(blob)

    def write_bootargs(self, args):
        base, _ = self._bootargs_span
        self.u.proxy.iface.writemem(base, args.to_bytes())

    def update_bootargs(self, keyval, logger=print):
        args = self.read_bootargs()
        old = deepcopy(args)
        args.update(keyval)
        self.write_bootargs(args)
        old.dump_diff(args, logger)



# SPDX-License-Identifier: MIT
import struct

from . import asm, sysreg
from .proxyutils import GuardedHeap

__all__ = ["dynamic_regs", "impdef_regs", "static_regs", "find_regs"]

def _all():
    for op1 in range(1 << 3):
        for CRn in (0b1011, 0b1111):
            for CRm in range(1 << 4):
                for op2 in range(1 << 3):
                    yield 3, op1, CRn, CRm, op2

dynamic_regs = [
    sysreg.ACNTVCT_EL0,
    sysreg.ACNTPCT_EL0,
]

impdef_regs = list(_all())
static_regs = [i for i in _all() if i not in dynamic_regs]

def find_regs(u, regs=None, block=1024, call=None, values=True):
    if regs is None:
        regs = impdef_regs

    p = u.proxy
    iface = u.iface

    data_len = 8 * block

    with GuardedHeap(u.heap) as heap:
        data_buffer = heap.malloc(data_len)

        template = asm.ARMAsm("""
            mov x2, x0
            mrs x2, s3_0_c0_c0_0
            str x2, [x1], #8
        """, 0x1000)

        mov, mrs, st = struct.unpack("3I", template.data)


        BAD = 0xacce5515abad1dea
        OOPS = 0xdeadc0dedeadc0de

        iregs = iter(regs)

        while True:
            insns = []
            bregs = []
            for i in iregs:
                op0, op1, CRn, CRm, op2 = enc = sysreg.sysreg_parse(i)
                bregs.append(enc)
                assert op0 == 3

                insns.extend((mov, mrs | (op1 << 16) | (CRn << 12) | (CRm << 8) | (op2 << 5), st))

                if len(bregs) >= block:
                    break

            if not bregs:
                break

            p.memset64(data_buffer, OOPS, data_len)
            u.exec(insns, BAD, data_buffer, call=call, silent=True, ignore_exceptions=True)
            data = iface.readmem(data_buffer, 8 * len(bregs))
            for reg, val in zip(bregs, struct.unpack(f"<{len(bregs)}Q", data)):
                if val == OOPS:
                    raise Exception(f"Failed to execute reg-finder code at {reg}")
                if val != BAD:
                    if values:
                        yield reg, val
                    else:
                        yield reg

if __name__ == "__main__":
    from m1n1.setup import *

    p.iodev_set_usage(IODEV.FB, 0)

    for reg, val in find_regs(u):
        print(f"{sysreg_name(reg)} ({', '.join(map(str, reg))}) = 0x{val:x}")

        try:
            u.msr(reg, val, silent=True)
        except:
            print(" - READONLY")
        try:
            u.mrs(reg, silent=True, call="el1")
        except:
            print(" - ### EL2 only ###")
        try:
            u.mrs(reg, silent=True, call="el0")
        except:
            pass
        else:
            print(" - *** EL0 accessible ***")



# SPDX-License-Identifier: MIT
import struct

from .base import *
from ...utils import *

## Syslog endpoint

class SyslogMessage(Register64):
    TYPE        = 59, 52

class Syslog_Init(SyslogMessage):
    TYPE        = 59, 52, Constant(8)
    ENTRYSIZE   = 39, 24
    COUNT       = 15, 0

class Syslog_GetBuf(SyslogMessage):
    TYPE        = 59, 52, Constant(1)
    SIZE        = 51, 44
    DVA         = 43, 0

class Syslog_Log(SyslogMessage):
    TYPE        = 59, 52, Constant(5)
    INDEX       = 7, 0

class ASCSysLogEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = SyslogMessage
    SHORT = "syslog"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.entrysize = None
        self.count = None
        self.iobuffer = None
        self.iobuffer_dva = None
        self.started = False

    @msg_handler(8, Syslog_Init)
    def Init(self, msg):
        self.entrysize = msg.ENTRYSIZE
        self.count = msg.COUNT
        self.log(f"count {self.count}, entrysize {self.entrysize}")
        return True

    @msg_handler(1, Syslog_GetBuf)
    def GetBuf(self, msg):
        size = align(0x1000 * msg.SIZE, 0x4000)

        if self.iobuffer:
            print("WARNING: trying to reset iobuffer!")

        if msg.DVA:
            self.iobuffer_dva = msg.DVA
            self.log(f"buf prealloc at dva {self.iobuffer_dva:#x}")
        else:
            self.iobuffer, self.iobuffer_dva = self.asc.ioalloc(size)
            self.log(f"buf {self.iobuffer:#x} / {self.iobuffer_dva:#x}")
            self.send(Syslog_GetBuf(SIZE=size // 0x1000, DVA=self.iobuffer_dva))

        self.started = True
        return True

    @msg_handler(5, Syslog_Log)
    def Log(self, msg):
        stride = 0x20 + self.entrysize
        log = self.asc.ioread(self.iobuffer_dva + msg.INDEX * stride, stride)
        hdr, unk, context, logmsg = struct.unpack(f"<II24s{self.entrysize}s", log)
        context = context.split(b"\x00")[0].decode("ascii")
        logmsg = logmsg.split(b"\x00")[0].decode("ascii").rstrip("\n")
        self.log(f"* [{context}]{logmsg}")
        self.send(msg)
        return True



# SPDX-License-Identifier: MIT
from ..utils import *
from enum import IntEnum


### NOTE: This is "MSR10j" (M1 Max), and there definitely *ARE* differences from M1

class R_IRQS(Register32):
    DONE = 0
    DBGSTS_ERROR = 1
    READ_ERROR = 3
    # This doesn't actually trigger on bad IOVAs?
    WRITE_ERROR_MAYBE = 4
    DECOMPRESSION_ERROR = 9
    CONTEXT_SWITCH = 10
    _BIT11 = 11
    AXI_ERROR = 12
    _BIT13 = 13


class E_ROTATE(IntEnum):
    # clockwise rotations
    ROT_0 = 0
    ROT_90 = 1
    ROT_180 = 2
    ROT_270 = 3


class R_FLIP_ROTATE(Register32):
    ROTATE = 1, 0, E_ROTATE
    FLIP_UPDOWN = 2
    FLIP_LEFTRIGHT = 3


class R_SCALE_FLAGS(Register32):
    EN = 0
    MAKE_THE_OUTPUT_YELLOW = 1
    # only when bit1 is set, only on H scaling
    MAKE_A_BLUE_LINE_APPEAR = 4


class ScalerMainRegs(RegMap):
    # on startup 1 will be written followed by 0
    # but it's not clear if that actually does anything
    HW_VERSION                      = 0x00000, Register32
    # bit0 = normal, bit1 = apiodma related
    # if things are reset here, reading all other regs will *HANG*
    RESET                           = 0x00004, Register32
    # can set 0x1f00
    RESET_APIODMA_RELATED           = 0x00008, Register32
    IS_RUNNING                      = 0x0000c, Register32
    # writable, can set to 0xfff
    REG_0x10                        = 0x00010, Register32
    REGISTER_FIFO_AVAILABILITY      = 0x00014, Register32
    # XNU sets 0x121b, we can at most set 0x3e1b
    IRQ_ENABLE                      = 0x00018, R_IRQS
    MSR_GLBL_IRQSTS                 = 0x0001c, R_IRQS
    FRAME_COUNT                     = 0x00020, Register32

    # can set 7
    REG_0x58                        = 0x00058, Register32

    # can set 0xffff
    REG_0x74                        = 0x00074, Register32

    # 1, or 3 if readonly??
    START                           = 0x00080, Register32
    # can set all bits
    REG_0x84                        = 0x00084, Register32
    # can set all bits
    REG_0x88                        = 0x00088, Register32
    # can set 0x8000ffff
    REG_0x8c                        = 0x0008c, Register32

    # can set all bits
    REG_0x98                        = 0x00098, Register32
    # 0x3f3d?
    MSR_CTRL_DBGSTS                 = 0x0009c, Register32
    # can set 3
    REG_0xa0                        = 0x000a0, Register32
    PROFILING_RELATED               = 0x000a4, Register32

    # Can set bits 0/1/2
    # Does something breaking horizontal scaling
    # bit2 seems to affect alpha output
    PIXEL_AVERAGING                 = 0x000e4, Register32

    TRANSFORM_ID                    = 0x00110, Register32

    RDMA_THING0                     = 0x00180, Register32
    RDMA_THING1                     = 0x00184, Register32
    RDMA_THING2                     = 0x00188, Register32
    RDMA_THING3                     = 0x0018c, Register32
    RDMA_THING4                     = 0x00190, Register32

    # there's probably another source plane existing?
    SRC_PLANE0_LO                   = 0x00198, Register32
    SRC_PLANE0_HI                   = 0x0019c, Register32
    SRC_PLANE1_LO                   = 0x001a0, Register32
    SRC_PLANE1_HI                   = 0x001a4, Register32
    SRC_PLANE2_LO                   = 0x001a8, Register32
    SRC_PLANE2_HI                   = 0x001ac, Register32

    SRC_PLANE0_COMPRESSEDTHING_LO   = 0x001b8, Register32
    SRC_PLANE0_COMPRESSEDTHING_HI   = 0x001bc, Register32
    SRC_PLANE1_COMPRESSEDTHING_LO   = 0x001c0, Register32
    SRC_PLANE1_COMPRESSEDTHING_HI   = 0x001c4, Register32
    SRC_PLANE2_COMPRESSEDTHING_LO   = 0x001c8, Register32
    SRC_PLANE2_COMPRESSEDTHING_HI   = 0x001cc, Register32

    SRC_PLANE0_STRIDE               = 0x001d8, Register32
    SRC_PLANE1_STRIDE               = 0x001dc, Register32
    SRC_PLANE2_STRIDE               = 0x001e0, Register32

    # seems to be in "pixels"
    SRC_PLANE0_OFFSET               = 0x001e8, Register32
    SRC_PLANE1_OFFSET               = 0x001ec, Register32
    SRC_PLANE2_OFFSET               = 0x001f0, Register32

    SRC_SWIZZLE                     = 0x001f8, Register32
    SRC_W                           = 0x001fc, Register32
    SRC_H                           = 0x00200, Register32
    CACHE_HINTS_THING0              = irange(0x00204, 4, 4), Register32
    CACHE_HINTS_THING1              = irange(0x00214, 4, 4), Register32
    TUNABLES_THING0                 = irange(0x00224, 4, 4), Register32
    SRC_SIZE_THING2                 = 0x00234, Register32
    SRC_SIZE_THING3                 = 0x00238, Register32
    SRC_SIZE_THING4                 = 0x0023c, Register32

    SRC_SIZE_THING5                 = 0x00244, Register32
    SRC_SIZE_THING6                 = 0x00248, Register32
    SRC_SIZE_THING7                 = 0x0024c, Register32

    WDMA_THING0                     = 0x00280, Register32
    WDMA_THING1                     = 0x00284, Register32
    WDMA_THING2                     = 0x00288, Register32
    WDMA_THING3                     = 0x0028c, Register32
    DST_PLANE0_LO                   = 0x00290, Register32
    DST_PLANE0_HI                   = 0x00294, Register32
    DST_PLANE1_LO                   = 0x00298, Register32
    DST_PLANE1_HI                   = 0x0029c, Register32
    DST_PLANE2_LO                   = 0x002a0, Register32
    DST_PLANE2_HI                   = 0x002a4, Register32
    DST_PLANE0_COMPRESSEDTHING_LO   = 0x002a8, Register32
    DST_PLANE0_COMPRESSEDTHING_HI   = 0x002ac, Register32
    DST_PLANE1_COMPRESSEDTHING_LO   = 0x002b0, Register32
    DST_PLANE1_COMPRESSEDTHING_HI   = 0x002b4, Register32
    DST_PLANE2_COMPRESSEDTHING_LO   = 0x002b8, Register32
    DST_PLANE2_COMPRESSEDTHING_HI   = 0x002bc, Register32
    DST_PLANE0_STRIDE               = 0x002c0, Register32
    DST_PLANE1_STRIDE               = 0x002c4, Register32
    DST_PLANE2_STRIDE               = 0x002c8, Register32
    DST_PLANE0_OFFSET               = 0x002cc, Register32
    DST_PLANE1_OFFSET               = 0x002d0, Register32
    DST_PLANE2_OFFSET               = 0x002d4, Register32
    DST_SWIZZLE                     = 0x002d8, Register32
    DST_W                           = 0x002dc, Register32
    DST_H                           = 0x002e0, Register32
    # uhh is there a macos bug with these? last val always overwritten
    CACHE_HINTS_THING2              = irange(0x002e4, 3, 4), Register32
    CACHE_HINTS_THING3              = irange(0x002f0, 3, 4), Register32
    TUNABLES_THING1                 = irange(0x002fc, 3, 4), Register32
    DST_SIZE_THING2                 = 0x00308, Register32
    DST_SIZE_THING3                 = 0x0030c, Register32
    DST_SIZE_THING4                 = 0x00310, Register32
    DST_SIZE_THING5                 = 0x00314, Register32
    DST_SIZE_THING6                 = 0x00318, Register32
    DST_SIZE_THING7                 = 0x0031c, Register32

    FLIP_ROTATE                     = 0x00380, R_FLIP_ROTATE

    # can set bit 0/1
    # the output obviously changes when this is set
    PSEUDO_LINEAR_SCALING           = 0x00480, Register32

    SCALE_V_FLAGS                   = 0x01000, R_SCALE_FLAGS
    # No idea what a DDA is? Q1.22 or U1.22 (23 bits total)
    # Also macOS doesn't touch a bunch of the V ones (uses H instead???)
    SCALE_V_DDA_THING0              = 0x01004, Register32
    SCALE_V_DDA_THING1              = 0x01008, Register32
    # Q4.22 or U4.22 (26 bits total)
    SCALE_V_RATIO_0                 = 0x0100c, Register32
    SCALE_V_RATIO_1                 = 0x01010, Register32
    SCALE_V_RATIO_2                 = 0x01014, Register32
    SCALE_V_RATIO_3                 = 0x01018, Register32
    SCALE_V_DDA_THING2              = 0x0101c, Register32
    SCALE_V_RATIO_4                 = 0x01020, Register32
    SCALE_V_RATIO_5                 = 0x01024, Register32

    # 9 taps, 32 phases polyphase resampling filter
    # Q4.12 (16-bit total) fixed point filter coeffs
    # packed into 32-bit registers, 3 sets of filters total (chroma/luma/alpha??)
    # exact ordering and arithmetic performed not yet clear
    SCALE_FILTER_V_BLOCK0           = irange(0x01400, 9 * 32, 4), Register32
    SCALE_FILTER_V_BLOCK1           = irange(0x01c00, 9 * 32 // 2, 4), Register32

    SCALE_H_FLAGS                   = 0x02000, R_SCALE_FLAGS
    # No idea what a DDA is? Q1.22 or U1.22 (23 bits total)
    SCALE_H_DDA_THING0              = 0x02004, Register32
    SCALE_H_DDA_THING1              = 0x02008, Register32
    # Q4.22 or U4.22 (26 bits total)
    SCALE_H_RATIO_0                 = 0x0200c, Register32
    SCALE_H_RATIO_1                 = 0x02010, Register32
    SCALE_H_RATIO_2                 = 0x02014, Register32
    SCALE_H_RATIO_3                 = 0x02018, Register32
    SCALE_H_DDA_THING2              = 0x0201c, Register32
    SCALE_H_RATIO_4                 = 0x02020, Register32
    SCALE_H_RATIO_5                 = 0x02024, Register32

    # 15 taps, 32 phases polyphase resampling filter
    SCALE_FILTER_H_BLOCK0           = irange(0x02400, 15 * 32, 4), Register32
    SCALE_FILTER_H_BLOCK1           = irange(0x02c00, 15 * 32 // 2, 4), Register32



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, argparse
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))
sys.path.append("/home/eileen/asahi/avd")  # git clone https://github.com/eiln/avd.git
# Decode with our own generated instruction stream

from m1n1.setup import *
from m1n1.utils import *
from m1n1.fw.avd import *
from tools.common import ffprobe

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    #
    # Usage:
    #     ffmpeg -i input.mp4 -c:v copy (or reencode libx264) input.264
    #     python3 experiments/avd.py -i input.264 -a
    #
    # - Supports .264, .265, and .ivf formats.
    # - Regarding the codec, it's whatever codec features are supported.
    #   Check avid for details.
    # - Also ensure to change the sys.path.append above to the avid repo
    #   as it does not install system-wide.
    #
    parser.add_argument('-i', '--input', type=str, required=True, help="path to input bitstream")
    parser.add_argument('-n', '--num', type=int, default=1, help="frame count")
    parser.add_argument('-a', '--all', action='store_true', help="run all frames")
    parser.add_argument('-x', '--stfu', action='store_true')
    parser.add_argument('-p', '--poll', action='store_true', help="poll iommu space")
    parser.add_argument('-d', '--display', action='store_true', help="display frames")
    parser.add_argument('--save-raw', type=str, default="", help="file name to save raw yuv")
    parser.add_argument('--save-images', type=str, default="", help="dirname to save images")
    args = parser.parse_args()
    mode = ffprobe(args.input)

    if   (mode == "h264"):
        from avid.h264.decoder import AVDH264Decoder
        dec = AVDH264Decoder()
    elif (mode == "h265"):
        from avid.h265.decoder import AVDH265Decoder
        dec = AVDH265Decoder()
    elif (mode == "vp09"):
        from avid.vp9.decoder import AVDVP9Decoder
        dec = AVDVP9Decoder()
    else:
        raise RuntimeError("unsupported codec")
    if (args.stfu):
        dec.stfu = True
        dec.hal.stfu = True
    nal_stop = 0 if args.all else 1
    units = dec.setup(args.input, nal_stop=nal_stop, num=args.num)

    avd = AVDDevice(u)
    if   (mode == "h264"):
        avd.decoder = AVDH264Dec(avd)
    elif (mode == "h265"):
        avd.decoder = AVDH265Dec(avd)
    elif (mode == "vp09"):
        avd.decoder = AVDVP9Dec(avd)
    else:
        raise RuntimeError("unsupported codec")
    avd.decoder.winname = args.input
    if (args.stfu):
        avd.stfu = True
    avd.boot()

    rawvideo = b''
    num = len(units) if args.all else min(args.num, len(units))
    for i,unit in enumerate(units[:num]):
        print(unit)
        inst = dec.decode(unit)
        if (i == 0):
            avd.ioalloc_at(0x0, dec.allocator_top(), stream=0, val=0)
            if (args.poll):
                avd.iomon.add(0x0, 0x100000)
        frame = avd.decoder.decode(dec.ctx, unit, inst, args.display)
        if (frame != None):
            if (args.save_raw):
                rawvideo += frame.y_data + frame.uv_data
            if (args.save_images):
                os.makedirs(f"data/out/{args.save_images}", exist_ok=True)
                path = os.path.join(f"data/out/{args.save_images}", "out%03d.png" % (avd.decoder.count))
                cv2.imwrite(path, frame.img)
    if (args.save_raw):
        path = os.path.join(f"data/out/{args.save_raw}")
        open(path, "wb").write(rawvideo)



# SPDX-License-Identifier: MIT

from m1n1.trace.dart import DARTTracer

DARTTracer = DARTTracer._reloadcls()

DEVICES = [
    "/arm-io/dart-pmp",
    "/arm-io/dart-sep",
    "/arm-io/dart-sio",
    "/arm-io/dart-usb1",
    "/arm-io/dart-disp0",
    "/arm-io/dart-dcp",
    "/arm-io/dart-dispext0",
    "/arm-io/dart-dcpext",
    "/arm-io/dart-scaler",
]

dart_tracers = {}

for i in DEVICES:
    tracer = DARTTracer(hv, i, verbose=3)
    tracer.start()

    dart_tracers[i.split("-")[-1]] = tracer

del tracer
